@@ -1,7 +1,7 @@
-package Sub::Exporter;
-
+use 5.006;
 use strict;
 use warnings;
+package Sub::Exporter;
 
 use Carp ();
 use Data::OptList ();
@@ -14,11 +14,11 @@ Sub::Exporter - a sophisticated exporter for custom-built routines
 
 =head1 VERSION
 
-version 0.976
+version 0.982
 
 =cut
 
-our $VERSION = '0.976';
+our $VERSION = '0.982';
 
 =head1 SYNOPSIS
 
@@ -28,7 +28,7 @@ Sub::Exporter must be used in two places.  First, in an exporting module:
   package Text::Tweaker;
   use Sub::Exporter -setup => {
     exports => [
-      qw(squish titlecase) # always works the same way
+      qw(squish titlecase), # always works the same way
       reformat => \&build_reformatter, # generator to build exported function
       trim     => \&build_trimmer,
       indent   => \&build_indenter,
@@ -176,9 +176,29 @@ The following keys are valid in C<%config>:
   groups  - a list of groups to provide for exporting; each must be followed by
             either (a) a list of exports, possibly with arguments for each
             export, or (b) a generator
+
   collectors - a list of names into which values are collected for use in
                routine generation; each name may be followed by a validator
 
+In addition to the basic options above, a few more advanced options may be
+passed:
+
+  into_level - how far up the caller stack to look for a target (default 0)
+  into       - an explicit target (package) into which to export routines
+
+In other words: Sub::Exporter installs a C<import> routine which, when called,
+exports routines to the calling namespace.  The C<into> and C<into_level>
+options change where those exported routines are installed.
+
+  generator  - a callback used to produce the code that will be installed
+               default: Sub::Exporter::default_generator
+
+  installer  - a callback used to install the code produced by the generator
+               default: Sub::Exporter::default_installer
+
+For information on how these callbacks are used, see the documentation for
+C<L</default_generator>> and C<L</default_installer>>.
+
 =head2 Export Configuration
 
 The C<exports> list may be provided as an array reference or a hash reference.
@@ -329,6 +349,11 @@ hashref containing the following entries:
   class       - the package on which the importer was called
   into        - the package into which exports will be exported
 
+Collectors with all-caps names (that is, made up of underscore or capital A
+through Z) are reserved for special use.  The only currently implemented
+special collector is C<INIT>, whose hook (if present in the exporter
+configuration) is always run before any other hook.
+
 =head1 CALLING THE EXPORTER
 
 Arguments to the exporter (that is, the arguments after the module name in a
@@ -363,11 +388,13 @@ first argument, in a hashref.
 
 These options are:
 
-  into_level - how far up the caller stack to look for a target (default 0)
-  into       - an explicit target (package) into which to export routines
+  into_level
+  into
+  generator
+  installer
 
-Providing both C<into_level> and C<into> will cause an exception to be
-thrown.
+These override the same-named configuration options described in L</EXPORTER
+CONFIGURATION>.
 
 =cut
 
@@ -459,7 +486,7 @@ sub _expand_group {
     # I'm not very happy with this code for hiding -prefix and -suffix, but
     # it's needed, and I'm not sure, offhand, how to make it better.
     # -- rjbs, 2006-12-05
-    my $group_arg = $group_arg ? { %$group_arg } : {};
+    my $group_arg = $merge ? { %$merge } : {};
     delete $group_arg->{-prefix};
     delete $group_arg->{-suffix};
 
@@ -484,19 +511,13 @@ sub _expand_group {
   }
 }
 
-# Given a config and pre-canonicalized importer args, remove collections from
-# the args and return them.
-sub _collect_collections {
-  my ($config, $import_args, $class, $into) = @_;
-  my %collection;
-
-  my @collections
-    = map  { splice @$import_args, $_, 1 }
-      grep { exists $config->{collectors}{ $import_args->[$_][0] } }
-      reverse 0 .. $#$import_args;
+sub _mk_collection_builder {
+  my ($col, $etc) = @_;
+  my ($config, $import_args, $class, $into) = @$etc;
 
   my %seen;
-  for my $collection (@collections) {
+  sub {
+    my ($collection) = @_;
     my ($name, $value) = @$collection;
 
     Carp::croak "collection $name provided multiple times in import"
@@ -519,10 +540,29 @@ sub _collect_collections {
       }
     }
 
-    $collection{ $name } = $value;
+    $col->{ $name } = $value;
+  }
+}
+
+# Given a config and pre-canonicalized importer args, remove collections from
+# the args and return them.
+sub _collect_collections {
+  my ($config, $import_args, $class, $into) = @_;
+
+  my @collections
+    = map  { splice @$import_args, $_, 1 }
+      grep { exists $config->{collectors}{ $import_args->[$_][0] } }
+      reverse 0 .. $#$import_args;
+
+  unshift @collections, [ INIT => {} ] if $config->{collectors}{INIT};
+
+  my $col = {};
+  my $builder = _mk_collection_builder($col, \@_);
+  for my $collection (@collections) {
+    $builder->($collection)
   }
 
-  return \%collection;
+  return $col;
 }
 
 =head1 SUBROUTINES
@@ -550,13 +590,10 @@ The exporter is built by C<L</build_exporter>>.
 
 =cut
 
-# \%special is for experimental options that may or may not be kept around and,
-# probably, moved to \%config.  These are also passed along to build_exporter.
-
 sub setup_exporter {
   my ($config)  = @_;
 
-  Carp::croak q(into and into_level may not both be supplied to exporter)
+  Carp::croak 'into and into_level may not both be supplied to exporter'
     if exists $config->{into} and exists $config->{into_level};
 
   my $as   = delete $config->{as}   || 'import';
@@ -599,7 +636,17 @@ my %valid_config_key;
 BEGIN {
   %valid_config_key =
     map { $_ => 1 }
-    qw(collectors exporter exports groups into into_level)
+    qw(as collectors installer generator exports groups into into_level),
+    qw(exporter), # deprecated
+}
+
+sub _assert_collector_names_ok {
+  my ($collectors) = @_;
+
+  for my $reserved_name (grep { /\A[_A-Z]+\z/ } keys %$collectors) {
+    Carp::croak "unknown reserved collector name: $reserved_name"
+      if $reserved_name ne 'INIT';
+  }
 }
 
 sub _rewrite_build_config {
@@ -609,6 +656,15 @@ sub _rewrite_build_config {
     Carp::croak "unknown options (@keys) passed to Sub::Exporter";
   }
 
+  Carp::croak q(into and into_level may not both be supplied to exporter)
+    if exists $config->{into} and exists $config->{into_level};
+
+  # XXX: Remove after deprecation period.
+  if ($config->{exporter}) {
+    Carp::cluck "'exporter' argument to build_exporter is deprecated. Use 'installer' instead; the semantics are identical.";
+    $config->{installer} = delete $config->{exporter};
+  }
+
   Carp::croak q(into and into_level may not both be supplied to exporter)
     if exists $config->{into} and exists $config->{into_level};
 
@@ -620,6 +676,8 @@ sub _rewrite_build_config {
     );
   }
 
+  _assert_collector_names_ok($config->{collectors});
+
   if (my @names = _key_intersection(@$config{qw(exports collectors)})) {
     Carp::croak "names (@names) used in both collections and exports";
   }
@@ -640,6 +698,9 @@ sub _rewrite_build_config {
 
   # by default, build an all-inclusive 'all' group
   $config->{groups}{all} ||= [ keys %{ $config->{exports} } ];
+
+  $config->{generator} ||= \&default_generator;
+  $config->{installer} ||= \&default_installer;
 }
 
 sub build_exporter {
@@ -655,6 +716,11 @@ sub build_exporter {
     Carp::croak q(into and into_level may not both be supplied to exporter)
       if exists $special->{into} and exists $special->{into_level};
 
+    if ($special->{exporter}) {
+      Carp::cluck "'exporter' special import argument is deprecated. Use 'installer' instead; the semantics are identical.";
+      $special->{installer} = delete $special->{exporter};
+    }
+
     my $into
       = defined $special->{into}       ? delete $special->{into}
       : defined $special->{into_level} ? caller(delete $special->{into_level})
@@ -662,9 +728,8 @@ sub build_exporter {
       : defined $config->{into_level}  ? caller($config->{into_level})
       :                                  caller(0);
 
-    my $export = delete $special->{exporter}
-              || $config->{exporter}
-              || \&default_exporter;
+    my $generator = delete $special->{generator} || $config->{generator};
+    my $installer = delete $special->{installer} || $config->{installer};
 
     # this builds a AOA, where the inner arrays are [ name => value_ref ]
     my $import_args = Data::OptList::mkopt([ @_ ]);
@@ -677,61 +742,67 @@ sub build_exporter {
     my $to_import = _expand_groups($class, $config, $import_args, $collection);
 
     # now, finally $import_arg is really the "to do" list
-    for (@$to_import) {
-      _do_import($class, @$_, $collection, $config, $into, $export);
-    }
+    _do_import(
+      {
+        class     => $class,
+        col       => $collection,
+        config    => $config,
+        into      => $into,
+        generator => $generator,
+        installer => $installer,
+      },
+      $to_import,
+    );
   };
 
   return $import;
 }
 
 sub _do_import {
-  my ($class, $name, $arg, $collection, $config, $into, $export) = @_;
+  my ($arg, $to_import) = @_;
 
-  my ($generator, $as);
+  my @todo;
 
-  if ($arg and Params::Util::_CODELIKE($arg)) { ## no critic
-    # This is the case when a group generator has inserted name/code pairs.
-    $generator = sub { $arg };
-    $as = $name;
-  } else {
-    $arg = { $arg ? %$arg : () };
+  for my $pair (@$to_import) {
+    my ($name, $import_arg) = @$pair;
 
-    Carp::croak qq("$name" is not exported by the $class module)
-      unless (exists $config->{exports}{$name});
+    my ($generator, $as);
 
-    $generator = $config->{exports}{$name};
+    if ($import_arg and Params::Util::_CODELIKE($import_arg)) { ## no critic
+      # This is the case when a group generator has inserted name/code pairs.
+      $generator = sub { $import_arg };
+      $as = $name;
+    } else {
+      $import_arg = { $import_arg ? %$import_arg : () };
 
-    $as = exists $arg->{-as} ? (delete $arg->{-as}) : $name;
+      Carp::croak qq("$name" is not exported by the $arg->{class} module)
+        unless exists $arg->{config}{exports}{$name};
+
+      $generator = $arg->{config}{exports}{$name};
+
+      $as = exists $import_arg->{-as} ? (delete $import_arg->{-as}) : $name;
+    }
+
+    my $code = $arg->{generator}->(
+      { 
+        class     => $arg->{class},
+        name      => $name,
+        arg       => $import_arg,
+        col       => $arg->{col},
+        generator => $generator,
+      }
+    );
+
+    push @todo, $as, $code;
   }
 
-  $export->($class, $generator, $name, $arg, $collection, $as, $into);
-}
-
-# XXX: Consider implementing a _export_args routine that takes the arguments to
-# _export and returns a hash of named params.  This lets other people write
-# exporters without tying me down to one set of @_ contents.  Maybe that's
-# premature guarantee, though, unless I guarantee that @_ will never get
-# /smaller/.
-
-=head2 default_exporter
-
-This is Sub::Exporter's default exporter.  It does what Sub::Exporter promises:
-it calls generators with the three normal arguments, then installs the code
-into the target package.
-
-B<Warning!>  Its interface isn't really stable yet, so don't rely on it.  It's
-only named here so that you can pass it in to the exporter builder.  It will
-have a stable interface in the future so that it may be more easily replaced.
-
-=cut
-
-sub default_exporter {
-  my ($class, $generator, $name, $arg, $collection, $as, $into) = @_;
-  _install(
-    _generate($class, $generator, $name, $arg, $collection),
-    $into,
-    $as,
+  $arg->{installer}->(
+    {
+      class => $arg->{class},
+      into  => $arg->{into},
+      col   => $arg->{col},
+    },
+    \@todo,
   );
 }
 
@@ -750,8 +821,28 @@ sub default_exporter {
 #   }
 # }
 
-sub _generate {
-  my ($class, $generator, $name, $arg, $collection) = @_;
+=head2 default_generator
+
+This is Sub::Exporter's default generator.  It takes bits of configuration that
+have been gathered during the import and turns them into a coderef that can be
+installed.
+
+  my $code = default_generator(\%arg);
+
+Passed arguments are:
+
+  class - the class on which the import method was called
+  name  - the name of the export being generated
+  arg   - the arguments to the generator
+  col   - the collections
+
+  generator - the generator to be used to build the export (code or scalar ref)
+
+=cut
+
+sub default_generator {
+  my ($arg) = @_;
+  my ($class, $name, $generator) = @$arg{qw(class name generator)};
 
   if (not defined $generator) {
     my $code = $class->can($name)
@@ -762,33 +853,81 @@ sub _generate {
   # I considered making this "$class->$generator(" but it seems that
   # overloading precedence would turn an overloaded-as-code generator object
   # into a string before code. -- rjbs, 2006-06-11
-  return $generator->($class, $name, $arg, $collection)
+  return $generator->($class, $name, $arg->{arg}, $arg->{col})
     if Params::Util::_CODELIKE($generator); ## no critic Private
 
   # This "must" be a scalar reference, to a generator method name.
   # -- rjbs, 2006-12-05
-  return $class->$$generator($name, $arg, $collection);
+  return $class->$$generator($name, $arg->{arg}, $arg->{col});
 }
 
-sub _install {
-  my ($code, $into, $as) = @_;
-  # Allow as isa ARRAY to push onto an array?
-  # Allow into isa HASH to install name=>code into hash?
+=head2 default_installer
 
-  if (ref $as eq 'SCALAR') {
-    $$as = $code;
-  } elsif (ref $as) {
-    Carp::croak "invalid reference type for $as: " . ref $as;
-  } else {
-    Sub::Install::reinstall_sub({ code => $code, into => $into, as => $as });
+This is Sub::Exporter's default installer.  It does what Sub::Exporter
+promises: it installs code into the target package.
+
+  default_installer(\%arg, \@to_export);
+
+Passed arguments are:
+
+  into - the package into which exports should be delivered
+
+C<@to_export> is a list of name/value pairs.  The default exporter assigns code
+(the values) to named slots (the names) in the given package.  If the name is a
+scalar reference, the scalar reference is made to point to the code reference
+instead.
+
+=cut
+
+sub default_installer {
+  my ($arg, $to_export) = @_;
+
+  for (my $i = 0; $i < @$to_export; $i += 2) {
+    my ($as, $code) = @$to_export[ $i, $i+1 ];
+
+    # Allow as isa ARRAY to push onto an array?
+    # Allow into isa HASH to install name=>code into hash?
+
+    if (ref $as eq 'SCALAR') {
+      $$as = $code;
+    } elsif (ref $as) {
+      Carp::croak "invalid reference type for $as: " . ref $as;
+    } else {
+      Sub::Install::reinstall_sub({
+        code => $code,
+        into => $arg->{into},
+        as   => $as
+      });
+    }
   }
 }
 
+sub default_exporter {
+  Carp::cluck "default_exporter is deprecated; call default_installer instead; the semantics are identical";
+  goto &default_installer;
+}
+
 =head1 EXPORTS
 
 Sub::Exporter also offers its own exports: the C<setup_exporter> and
 C<build_exporter> routines described above.  It also provides a special "setup"
-group, which will setup an exporter using the parameters passed to it.
+collector, which will set up an exporter using the parameters passed to it.
+
+Note that the "setup" collector (seen in examples like the L</SYNOPSIS> above)
+uses C<build_exporter>, not C<setup_exporter>.  This means that the special
+arguments like "into" and "as" for C<setup_exporter> are not accepted here.
+Instead, you may write something like:
+
+  use Sub::Exporter
+    { into => 'Target::Package' },
+    -setup => {
+      -as     => 'do_import',
+      exports => [ ... ],
+    }
+  ;
+
+Finding a good reason for wanting to do this is left as as exercise for the
+reader.
 
 =cut
 
